
#ifndef PORYGONLANG_TABLEEVALVALUE_HPP
#define PORYGONLANG_TABLEEVALVALUE_HPP
#include <utility>
#include <map>
#include "EvalValue.hpp"
#include "../../Utilities/Random.hpp"
#include "../EvalValuePointer.hpp"

using namespace std;

namespace Porygon::Evaluation {
    class TableEvalValue : public EvalValue {
    protected:
        explicit TableEvalValue(shared_ptr<map<Utilities::HashedString, EvalValuePointer>>  table, size_t hash)
        :  _table(std::move(table)),
           _hash(hash)
        {
        }

        const shared_ptr<map<Utilities::HashedString, EvalValuePointer>> _table;
        const size_t _hash;
    public:
        explicit TableEvalValue(shared_ptr<map<Utilities::HashedString, EvalValuePointer>> table) :
            _table(std::move(table)),
            _hash(Utilities::Random::Get())
        {
        }

        [[nodiscard]]
        inline TypeClass GetTypeClass() const final {
            return TypeClass::Table;
        }

        [[nodiscard]]
        inline size_t GetHashCode() const final {
            return _hash;
        }

        [[nodiscard]]
        inline bool operator==(const EvalValue *b) const final {
            return this->_hash == b->GetHashCode();
        }

        bool operator!=(const EvalValue *b) const override {
            return !operator==(b);
        }

        [[nodiscard]]
        inline EvalValue* Clone() const override {
            return new TableEvalValue(_table, _hash);
        }

        [[nodiscard]]
        inline EvalValue* IndexValue(const EvalValue *val) const final {
            const auto stringKey = val->GetHashCode();
            return this->_table->at(Utilities::HashedString::CreateLookup(stringKey))->Clone();
        }

        [[nodiscard]]
        inline EvalValue* IndexValue(const Utilities::HashedString* hash) const final {
            return this->_table->at(*hash)->Clone();
        }

        [[nodiscard]]
        inline const EvalValue* IndexValue(const char *val) const {
            auto hash = Utilities::HashedString::ConstHash(val);
            return this->_table->at(Utilities::HashedString::CreateLookup(hash)).Clone();
        }

        inline void SetIndexValue(const EvalValue *key, const EvalValue* value) const final {
            auto hash = key->GetHashCode();
            auto lookup = Utilities::HashedString::CreateLookup(hash);
            auto insert = _table->insert({lookup, value});
            if (!insert.second) {
                _table->at(lookup).ClearAssign(value);
            }
        }

        [[nodiscard]]
        Iterator * GetKeyIterator() const override;

        [[nodiscard]]
        inline _Rb_tree_const_iterator<pair<const Utilities::HashedString, EvalValuePointer>> GetTableIterator() const{
            return _table->cbegin();
        };
        [[nodiscard]]
        inline _Rb_tree_const_iterator<pair<const Utilities::HashedString, EvalValuePointer>> GetTableIteratorEnd() const{
            return _table->cend();
        }

        [[nodiscard]] EvalValue *UnaryOperation(Binder::BoundUnaryOperation operation) const override;

        void SetIndexValue(const Utilities::HashedString *key, const EvalValue *value) const override;
    };

    class NumericTableEvalValue : public TableEvalValue{

        explicit NumericTableEvalValue(shared_ptr<map<Utilities::HashedString, EvalValuePointer>>  table, size_t hash)
        :  TableEvalValue(std::move(table), hash)
        {
        }
    public:
        explicit NumericTableEvalValue(shared_ptr<map<Utilities::HashedString, EvalValuePointer>> table)
         :TableEvalValue(std::move(table))
        {
        }

        [[nodiscard]] Iterator *GetKeyIterator() const final;

        [[nodiscard]] EvalValue *Clone() const override;
    };
}

#endif //PORYGONLANG_TABLEEVALVALUE_HPP
