
#ifndef PORYGONLANG_PARSEDEXPRESSION_HPP
#define PORYGONLANG_PARSEDEXPRESSION_HPP

#include <utility>
#include <memory>

#include "../Token.hpp"
#include "../UnaryOperatorKind.hpp"
#include "../BinaryOperatorKind.hpp"
#include "../../Utilities/HashedString.hpp"

namespace Porygon::Parser {
    enum class ParsedExpressionKind : uint8_t {
        Bad,

        LiteralInteger,
        LiteralFloat,
        LiteralString,
        LiteralBool,
        Nil,
        Variable,

        Unary,
        Binary,
        Parenthesized,
        FunctionCall,
        Indexer,
        PeriodIndexer,
        NumericalTable,
        Table,
    };

    class ParsedExpression {
        const unsigned int _position;
        const unsigned int _length;
    public:
        ParsedExpression(unsigned int position, unsigned int length)
                : _position(position),
                  _length(length) {
        }

        virtual ~ParsedExpression() = default;

        [[nodiscard]] virtual ParsedExpressionKind GetKind() const = 0;

        [[nodiscard]] inline unsigned int GetStartPosition() const {
            return _position;
        }

        [[nodiscard]] inline unsigned int GetEndPosition() const {
            return _position + _length - 1;
        }

        [[nodiscard]] inline unsigned int GetLength() const {
            return _length;
        }
    };

    class BadExpression : public ParsedExpression {
    public:
        BadExpression(unsigned int position, unsigned int length) : ParsedExpression(position, length) {}

        [[nodiscard]] inline ParsedExpressionKind GetKind() const final {
            return ParsedExpressionKind::Bad;
        }
    };

    class LiteralIntegerExpression : public ParsedExpression {
        const int64_t  _value;
    public:
        [[nodiscard]] inline ParsedExpressionKind GetKind() const final {
            return ParsedExpressionKind::LiteralInteger;
        }

        explicit LiteralIntegerExpression(const IntegerToken *token)
                : ParsedExpression(token->GetStartPosition(), token->GetLength()),
                  _value(token->GetValue()) {
        }

        [[nodiscard]] inline int64_t  GetValue() const {
            return _value;
        }
    };

    class LiteralFloatExpression : public ParsedExpression {
        const double _value;
    public:
        [[nodiscard]] inline ParsedExpressionKind GetKind() const final {
            return ParsedExpressionKind::LiteralFloat;
        }

        explicit LiteralFloatExpression(const FloatToken *token)
                : ParsedExpression(token->GetStartPosition(), token->GetLength()),
                  _value(token->GetValue()) {
        }

        [[nodiscard]] inline double GetValue() const {
            return _value;
        }
    };

    class LiteralStringExpression : public ParsedExpression {
        const u16string _value;
    public:
        [[nodiscard]] inline ParsedExpressionKind GetKind() const final {
            return ParsedExpressionKind::LiteralString;
        }

        explicit LiteralStringExpression(const StringToken *token)
                : ParsedExpression(token->GetStartPosition(), token->GetLength()),
                  _value(token->GetValue()) {
        }

        [[nodiscard]] inline const u16string &GetValue() const {
            return _value;
        }
    };

    class LiteralBoolExpression : public ParsedExpression {
        const bool _value;
    public:
        [[nodiscard]]inline ParsedExpressionKind GetKind() const final {
            return ParsedExpressionKind::LiteralBool;
        }

        explicit LiteralBoolExpression(const Token *token)
                : ParsedExpression(token->GetStartPosition(), token->GetLength()),
                  _value(token->GetKind() == TokenKind::TrueKeyword) {
        }

        [[nodiscard]]inline bool GetValue() const {
            return _value;
        }
    };

    class NilExpression : public ParsedExpression {
    public:
        [[nodiscard]] inline ParsedExpressionKind GetKind() const final {
            return ParsedExpressionKind::Nil;
        }

        explicit NilExpression(const Token *token)
                : ParsedExpression(token->GetStartPosition(), token->GetLength()) {
        }
    };

    class VariableExpression : public ParsedExpression {
        const HashedString _value;
    public:
        [[nodiscard]]inline ParsedExpressionKind GetKind() const final {
            return ParsedExpressionKind::Variable;
        }

        explicit VariableExpression(const IdentifierToken *token) : ParsedExpression(token->GetStartPosition(),
                                                                               token->GetLength()),
                                                              _value(HashedString(token->GetValue())) {
        }

        [[nodiscard]]inline HashedString GetValue() const {
            return _value;
        }
    };


    class ParenthesizedExpression : public ParsedExpression {
        const ParsedExpression *_expression;
    public:
        ~ParenthesizedExpression() override {
            delete _expression;
        }

        [[nodiscard]] inline ParsedExpressionKind GetKind() const final {
            return ParsedExpressionKind::Parenthesized;
        }

        explicit ParenthesizedExpression(ParsedExpression *innerExpression, unsigned int start, unsigned int length)
                : ParsedExpression(start, length), _expression(innerExpression) {
        }

        [[nodiscard]]inline const ParsedExpression *GetInnerExpression() const {
            return _expression;
        }
    };

    class UnaryExpression : public ParsedExpression {
        const UnaryOperatorKind _kind;
        const ParsedExpression *_operand;
    public:
        ~UnaryExpression() override {
            delete _operand;
        }

        [[nodiscard]]inline ParsedExpressionKind GetKind() const final {
            return ParsedExpressionKind::Unary;
        }

        UnaryExpression(UnaryOperatorKind kind, ParsedExpression *operand, unsigned int start, unsigned int length)
                : ParsedExpression(start, length),
                  _kind(kind), _operand(operand) {
        }

        [[nodiscard]]inline UnaryOperatorKind GetOperatorKind() const {
            return _kind;
        }

        [[nodiscard]]inline const ParsedExpression *GetOperand() const {
            return _operand;
        }
    };

    class BinaryExpression : public ParsedExpression {
        const BinaryOperatorKind _kind;
        const ParsedExpression *_left;
        const ParsedExpression *_right;
    public:
        ~BinaryExpression() override {
            delete _left;
            delete _right;
        }

        [[nodiscard]]inline ParsedExpressionKind GetKind() const final {
            return ParsedExpressionKind::Binary;
        }

        BinaryExpression(BinaryOperatorKind kind, ParsedExpression *left, ParsedExpression *right, unsigned int start,
                         unsigned int length)
                : ParsedExpression(start, length),
                  _kind(kind), _left(left), _right(right) {
        }

        [[nodiscard]]inline BinaryOperatorKind GetOperatorKind() const {
            return _kind;
        }

        [[nodiscard]]inline const ParsedExpression *GetLeft() const {
            return _left;
        }

        [[nodiscard]]inline const ParsedExpression *GetRight() const {
            return _right;
        }
    };

    class FunctionCallExpression : public ParsedExpression {
        const std::unique_ptr<ParsedExpression> _function;
        const vector<const ParsedExpression *> _parameters;
    public:
        FunctionCallExpression(ParsedExpression *function, vector<const ParsedExpression *> parameters,
                               unsigned int start, unsigned int length)
                : ParsedExpression(start, length),
                  _function(function), _parameters(std::move(parameters)) {
        }

        ~FunctionCallExpression() final {
            for (auto p : _parameters) {
                delete p;
            }
        }

        [[nodiscard]] inline ParsedExpressionKind GetKind() const final {
            return ParsedExpressionKind::FunctionCall;
        }

        [[nodiscard]] inline const ParsedExpression *GetFunction() const {
            return _function.get();
        }

        [[nodiscard]] inline const vector<const ParsedExpression *> *GetParameters() const {
            return &_parameters;
        }
    };

    class IndexExpression : public ParsedExpression {
        const ParsedExpression *_indexerExpression;
        const ParsedExpression *_indexExpression;
    public:
        IndexExpression(ParsedExpression *indexer, ParsedExpression *index, unsigned int start, unsigned int length)
                : ParsedExpression(start, length),
                  _indexerExpression(indexer), _indexExpression(index) {
        }

        ~IndexExpression() final {
            delete _indexerExpression;
            delete _indexExpression;
        }

        [[nodiscard]] inline ParsedExpressionKind GetKind() const final {
            return ParsedExpressionKind::Indexer;
        }

        [[nodiscard]] inline const ParsedExpression *GetIndexer() const {
            return _indexerExpression;
        }

        [[nodiscard]] inline const ParsedExpression *GetIndex() const {
            return _indexExpression;
        }
    };

    class PeriodIndexExpression : public ParsedExpression {
        const ParsedExpression *_indexerExpression;
        const HashedString _index;
    public:
        PeriodIndexExpression(ParsedExpression *indexer, const HashedString& index, unsigned int start, unsigned int length)
                : ParsedExpression(start, length),
                  _indexerExpression(indexer), _index(index) {
        }

        ~PeriodIndexExpression() final {
            delete _indexerExpression;
        }

        [[nodiscard]] inline ParsedExpressionKind GetKind() const final {
            return ParsedExpressionKind::PeriodIndexer;
        }

        [[nodiscard]] inline const ParsedExpression *GetIndexer() const {
            return _indexerExpression;
        }

        [[nodiscard]] inline const HashedString &GetIndex() const {
            return _index;
        }
    };


    class ParsedNumericalTableExpression : public ParsedExpression {
        const vector<const ParsedExpression *> _expressions;
    public:
        ParsedNumericalTableExpression(vector<const ParsedExpression *> expressions, unsigned int start,
                                       unsigned int length)
                : ParsedExpression(start, length), _expressions(std::move(expressions)) {
        }

        ~ParsedNumericalTableExpression() final {
            for (auto s: _expressions) {
                delete s;
            }
        }

        [[nodiscard]] inline const vector<const ParsedExpression *> *GetExpressions() const {
            return &_expressions;
        }

        [[nodiscard]] inline ParsedExpressionKind GetKind() const final {
            return ParsedExpressionKind::NumericalTable;
        }
    };
}

#endif //PORYGONLANG_PARSEDEXPRESSION_HPP
