#ifndef PORYGONLANG_USERDATA_HPP
#define PORYGONLANG_USERDATA_HPP

#include <utility>
#include <unordered_map>
#include <mutex>
#include "UserDataField.hpp"
#include "UserDataOperation.hpp"
#include "../Parser/BinaryOperatorKind.hpp"
#include "../Exception.hpp"

namespace Porygon::UserData {
    class UserData {
        Utilities::HashedString* _hashedString;
        std::unordered_map<uint32_t, unique_ptr<UserDataField>> _fields;
        std::mutex _mutex;

        // Binary operations
        std::vector<const UserDataBinaryOperation*> _addition = {};
        std::vector<const UserDataBinaryOperation*> _subtraction = {};
        std::vector<const UserDataBinaryOperation*> _multiplication = {};
        std::vector<const UserDataBinaryOperation*> _division = {};
        std::vector<const UserDataBinaryOperation*> _equality = {};
        std::vector<const UserDataBinaryOperation*> _inequality = {};
        std::vector<const UserDataBinaryOperation*> _lessThan = {};
        std::vector<const UserDataBinaryOperation*> _lessThanEquals = {};
        std::vector<const UserDataBinaryOperation*> _greaterThan = {};
        std::vector<const UserDataBinaryOperation*> _greaterThanEquals = {};

        bool (*_isCastable)(const ScriptType* type, bool explicitCast);
        Evaluation::EvalValue* (*_cast)(void* obj, const ScriptType* castType);

    public:
        explicit UserData(Utilities::HashedString* hashedString, const std::unordered_map<uint32_t, UserDataField *>& fields)
            : _hashedString(hashedString), _isCastable(nullptr), _cast(nullptr)
        {
            for (auto f: fields){
                _fields.insert({f.first, unique_ptr<UserDataField>(f.second)});
            }
        }

        ~UserData(){
            _fields.clear();
            for (auto o: _addition) delete o;
            for (auto o: _subtraction) delete o;
            for (auto o: _multiplication) delete o;
            for (auto o: _division) delete o;
            for (auto o: _equality) delete o;
            for (auto o: _inequality) delete o;
            for (auto o: _lessThan) delete o;
            for (auto o: _lessThanEquals) delete o;
            for (auto o: _greaterThan) delete o;
            for (auto o: _greaterThanEquals) delete o;

            delete _hashedString;
        }

        Utilities::HashedString* GetIdentifier(){
            return _hashedString;
        }

        [[nodiscard]]
        inline bool ContainsField(uint32_t fieldId) const{
            return _fields.find(fieldId) != _fields.end();
        }

        [[nodiscard]]
        inline UserDataField *GetField(uint32_t fieldId) const {
            return _fields.at(fieldId).get();
        }

        inline void CreateField(uint32_t fieldId, UserDataField *field) {
            std::lock_guard<std::mutex> guard(_mutex);
            _fields.insert({fieldId, unique_ptr<UserDataField>(field)});
        }

        [[nodiscard]]
        inline int32_t GetFieldCount() const{
            return _fields.size();
        }

        void SetIsCastable(bool (*cast)(const ScriptType* type, bool explicitCast)){
            _isCastable = cast;
        }

        bool IsCastable(const shared_ptr<const ScriptType>& castType, bool explicitCast) const{
            if (_isCastable == nullptr) return false;
            return _isCastable(castType.get(), explicitCast);
        }

        void SetCastFunc(Evaluation::EvalValue* (*cast)(void* obj, const ScriptType* castType)){
            _cast = cast;
        }

        Evaluation::EvalValue* Cast(void* obj, const ScriptType* castType) const{
            return _cast(obj, castType);
        }

        void AddBinaryOperation(Binder::BoundBinaryOperation kind, const UserDataBinaryOperation* op){
            switch (kind){
                case Binder::BoundBinaryOperation::Addition: return _addition.push_back(op);
                case Binder::BoundBinaryOperation::Subtraction:  return _subtraction.push_back(op);
                case Binder::BoundBinaryOperation::Multiplication:  return _multiplication.push_back(op);
                case Binder::BoundBinaryOperation::Division:  return _division.push_back(op);
                case Binder::BoundBinaryOperation::Equality:  return _equality.push_back(op);
                case Binder::BoundBinaryOperation::Inequality:  return _inequality.push_back(op);
                case Binder::BoundBinaryOperation::LessThan:  return _lessThan.push_back(op);
                case Binder::BoundBinaryOperation::LessThanEquals:  return _lessThanEquals.push_back(op);
                case Binder::BoundBinaryOperation::GreaterThan:  return _greaterThan.push_back(op);
                case Binder::BoundBinaryOperation::GreaterThanEquals:  return _greaterThanEquals.push_back(op);
                default:
                    stringstream err;
                    err << "Unknown binary operation kind: " << static_cast<int>(kind);
                    throw Exception(err.str().c_str());
            }
        }

    private:
        static const UserDataBinaryOperation* ResolveBinaryOperation(const vector<const UserDataBinaryOperation*>& v,
                const shared_ptr<const ScriptType>& b){
            for (auto o: v){
                if (o->IsValid(b)){
                    return o;
                }
            }
            return nullptr;
        }

    public:
        [[nodiscard]]
        const UserDataBinaryOperation* GetBinaryOperation(Parser::BinaryOperatorKind op, const shared_ptr<const ScriptType>& b){
            switch (op){

                case Parser::BinaryOperatorKind::Addition: return ResolveBinaryOperation(_addition, b);
                case Parser::BinaryOperatorKind::Subtraction: return ResolveBinaryOperation(_subtraction, b);
                case Parser::BinaryOperatorKind::Multiplication: return ResolveBinaryOperation(_multiplication, b);
                case Parser::BinaryOperatorKind::Division: return ResolveBinaryOperation(_division, b);
                case Parser::BinaryOperatorKind::Equality: return ResolveBinaryOperation(_equality, b);
                case Parser::BinaryOperatorKind::Inequality: return ResolveBinaryOperation(_inequality, b);
                case Parser::BinaryOperatorKind::Less: return ResolveBinaryOperation(_lessThan, b);
                case Parser::BinaryOperatorKind::LessOrEquals: return ResolveBinaryOperation(_lessThanEquals, b);
                case Parser::BinaryOperatorKind::Greater: return ResolveBinaryOperation(_greaterThan, b);
                case Parser::BinaryOperatorKind::GreaterOrEquals: return ResolveBinaryOperation(_greaterThanEquals, b);
                default: throw exception();
            }
        }
    };
}


#endif //PORYGONLANG_USERDATA_HPP
